use std::{
    io,
    path::{Path, PathBuf},
    sync::{Arc, Mutex},
};

use libarchive::{
    write::{Builder, FileWriter, WriteEntry},
    Entry, WriteFilter, WriteFormat,
};

/// Struct to abstract away the intrinsics of writing entries to an archive file
pub struct RepoArchiveWriter {
    ar: Arc<Mutex<FileWriter>>,
}

impl RepoArchiveWriter {
    pub async fn open<P: AsRef<Path>>(path: P) -> io::Result<Self> {
        let path = PathBuf::from(path.as_ref());

        // Open the archive file
        let ar = tokio::task::spawn_blocking(move || {
            let mut builder = Builder::new();
            builder.add_filter(WriteFilter::Gzip)?;
            builder.set_format(WriteFormat::PaxRestricted)?;

            builder.open_file(path)
        })
        .await
        .unwrap()?;

        Ok(Self {
            // In practice, mutex is only ever used by one thread at a time. It's simply here so we
            // can use spawn_blocking without issues.
            ar: Arc::new(Mutex::new(ar)),
        })
    }

    /// Add either a "desc" or "files" entry to the archive
    pub async fn add_entry<P: AsRef<Path>>(
        &self,
        full_name: &str,
        path: P,
        desc: bool,
    ) -> io::Result<()> {
        let metadata = tokio::fs::metadata(&path).await?;
        let file_size = metadata.len();

        let ar = Arc::clone(&self.ar);
        let full_name = String::from(full_name);
        let path = PathBuf::from(path.as_ref());

        Ok(tokio::task::spawn_blocking(move || {
            let mut ar_entry = WriteEntry::new();
            ar_entry.set_filetype(libarchive::archive::FileType::RegularFile);

            let name = if desc { "desc" } else { "files" };

            ar_entry.set_pathname(PathBuf::from(full_name).join(name));
            ar_entry.set_mode(0o100644);
            ar_entry.set_size(file_size.try_into().unwrap());

            ar.lock().unwrap().append_path(&mut ar_entry, path)
        })
        .await
        .unwrap()?)
    }

    pub async fn close(&self) -> io::Result<()> {
        let ar = Arc::clone(&self.ar);

        Ok(
            tokio::task::spawn_blocking(move || ar.lock().unwrap().close())
                .await
                .unwrap()?,
        )
    }
}
